use std::fmt::Display;

use serde::{Deserialize, Serialize};

#[derive(PartialEq, PartialOrd, Eq, Ord, Clone, Copy, Serialize, Deserialize)]
#[repr(u8)]
pub enum CardIndex {
    A = 14,
    K = 13,
    Q = 12,
    J = 11,
    N10 = 10,
    N9 = 9,
    N8 = 8,
    N7 = 7,
    N6 = 6,
    N5 = 5,
    N4 = 4,
    N3 = 3,
    N2 = 2,
}

impl CardIndex {
    pub fn get_blackjack_value(&self, ace_as_1: bool) -> u32 {
        match self {
            Self::J | Self::K | Self::Q => 10,
            Self::A => {
                if ace_as_1 {
                    1
                } else {
                    11
                }
            }
            _ => (*self as u8) as u32,
        }
    }
}

impl TryFrom<char> for CardIndex {
    type Error = &'static str;

    fn try_from(value: char) -> Result<Self, Self::Error> {
        match value {
            'A' => Ok(CardIndex::A),
            'K' => Ok(CardIndex::K),
            'Q' => Ok(CardIndex::Q),
            'J' => Ok(CardIndex::J),
            'T' => Ok(CardIndex::N10),
            '9' => Ok(CardIndex::N9),
            '8' => Ok(CardIndex::N8),
            '7' => Ok(CardIndex::N7),
            '6' => Ok(CardIndex::N6),
            '5' => Ok(CardIndex::N5),
            '4' => Ok(CardIndex::N4),
            '3' => Ok(CardIndex::N3),
            '2' => Ok(CardIndex::N2),
            _ => Err("Not a card"),
        }
    }
}

impl Display for CardIndex {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let index = match self {
            Self::A => "A".to_owned(),
            Self::K => "K".to_owned(),
            Self::Q => "Q".to_owned(),
            Self::J => "J".to_owned(),
            _ => (*self as u8).to_string(),
        };

        f.write_str(&index)?;

        Ok(())
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn blackjack_value_ace() {
        assert_eq!(CardIndex::A.get_blackjack_value(true), 1);
        assert_eq!(CardIndex::A.get_blackjack_value(false), 11);
    }

    #[test]
    fn blackjack_value_face() {
        assert_eq!(CardIndex::J.get_blackjack_value(true), 10);
        assert_eq!(CardIndex::Q.get_blackjack_value(true), 10);
        assert_eq!(CardIndex::K.get_blackjack_value(true), 10);
    }

    #[test]
    fn blackjack_value_num() {
        assert_eq!(CardIndex::N10.get_blackjack_value(true), 10);
        assert_eq!(CardIndex::N9.get_blackjack_value(true), 9);
        assert_eq!(CardIndex::N8.get_blackjack_value(true), 8);
        assert_eq!(CardIndex::N5.get_blackjack_value(true), 5);
        assert_eq!(CardIndex::N3.get_blackjack_value(true), 3);
        assert_eq!(CardIndex::N2.get_blackjack_value(true), 2);
    }
}
