use rand::{seq::SliceRandom, thread_rng};
use serde::{Deserialize, Deserializer, Serialize, Serializer};
use std::fmt::Display;

use super::{card::Card, card_index::CardIndex};

#[derive(Clone)]
pub struct Hand {
    cards: Vec<Card>,
}

impl Hand {
    pub fn new() -> Self {
        Hand { cards: Vec::new() }
    }

    pub fn add_card(&mut self, card: Card) {
        self.cards.push(card);
    }

    pub fn get_card(&self, i: usize) -> Option<&Card> {
        return self.cards.get(i);
    }

    pub fn count(&self) -> usize {
        self.cards.len()
    }

    pub fn remove_card(&mut self, i: usize) -> Card {
        self.cards.remove(i)
    }

    pub fn pop_card(&mut self) -> Option<Card> {
        self.cards.pop()
    }

    pub fn shuffle(&mut self) {
        self.cards.shuffle(&mut thread_rng());
    }

    pub fn is_backjack(&self) -> bool {
        self.cards.len() == 2 && self.get_blackjack_value() == 21
    }

    /**
     * Returns true if this hand could be split in a blackjack game.
     */
    pub fn is_valid_for_bj_split(&self) -> bool {
        if self.cards.len() != 2 {
            return false;
        }

        if let Some(card_0) = self.get_card(0) {
            if let Some(card_1) = self.get_card(1) {
                if card_0.index.get_blackjack_value(true) != card_1.index.get_blackjack_value(true)
                {
                    // Cards are not the same value
                    return false;
                }
            }
        }

        true
    }

    /**
     * Put another hand into this.
     * Leaves the other hand empty.
     */
    pub fn merge(&mut self, other: &mut Hand) {
        self.cards.append(&mut other.cards);
    }

    pub fn get_blackjack_value(&self) -> u32 {
        let mut sum: u32 = self
            .cards
            .iter()
            .map(|card| card.index.get_blackjack_value(true))
            .sum();

        // Add 10 for the all aces (adds up to 11) until we go over 21
        let aces = self
            .cards
            .iter()
            .filter(|card| card.index == CardIndex::A)
            .count();

        for _ in 0..aces {
            if sum + 10 > 21 {
                return sum;
            }

            sum += 10;
        }

        sum
    }
}

impl Display for Hand {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        for card in self.cards.iter() {
            card.fmt(f)?;
            f.write_str(" ")?;
        }

        Ok(())
    }
}

impl Serialize for Hand {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        self.cards.serialize(serializer)
    }
}

impl<'de> Deserialize<'de> for Hand {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let cards = Vec::<Card>::deserialize(deserializer)?;
        Ok(Hand { cards })
    }
}

#[cfg(test)]
mod tests {
    use super::super::card_suit::CardSuit;
    use super::*;

    #[test]
    fn is_blackjack_ace_first() {
        let mut hand = Hand::new();
        hand.add_card(Card {
            suit: CardSuit::Hearts,
            index: CardIndex::A,
        });
        hand.add_card(Card {
            suit: CardSuit::Diamonds,
            index: CardIndex::N10,
        });
        assert!(hand.is_backjack());
    }

    #[test]
    fn is_blackjack_ace_last() {
        let mut hand = Hand::new();
        hand.add_card(Card {
            suit: CardSuit::Diamonds,
            index: CardIndex::J,
        });
        hand.add_card(Card {
            suit: CardSuit::Hearts,
            index: CardIndex::A,
        });
        assert!(hand.is_backjack());
    }

    #[test]
    fn is_not_blackjack_too_many() {
        let mut hand = Hand::new();
        hand.add_card(Card {
            suit: CardSuit::Diamonds,
            index: CardIndex::J,
        });
        hand.add_card(Card {
            suit: CardSuit::Hearts,
            index: CardIndex::A,
        });
        hand.add_card(Card {
            suit: CardSuit::Spades,
            index: CardIndex::A,
        });
        assert!(!hand.is_backjack());
    }

    #[test]
    fn is_not_blackjack_too_few() {
        let mut hand = Hand::new();
        hand.add_card(Card {
            suit: CardSuit::Hearts,
            index: CardIndex::A,
        });
        assert!(!hand.is_backjack());
    }

    #[test]
    fn is_not_blackjack_value_21() {
        let mut hand = Hand::new();
        hand.add_card(Card {
            suit: CardSuit::Hearts,
            index: CardIndex::K,
        });
        hand.add_card(Card {
            suit: CardSuit::Clubs,
            index: CardIndex::N7,
        });
        hand.add_card(Card {
            suit: CardSuit::Hearts,
            index: CardIndex::N4,
        });
        assert!(!hand.is_backjack());
    }

    #[test]
    fn blackjack_value() {
        let mut hand = Hand::new();
        hand.add_card(Card {
            suit: CardSuit::Hearts,
            index: CardIndex::K,
        });
        assert_eq!(hand.get_blackjack_value(), 10);
    }

    #[test]
    fn merge_hands() {
        let mut hand = Hand::new();
        hand.add_card(Card {
            suit: CardSuit::Diamonds,
            index: CardIndex::N5,
        });
        hand.add_card(Card {
            suit: CardSuit::Spades,
            index: CardIndex::Q,
        });

        let mut other = Hand::new();
        hand.add_card(Card {
            suit: CardSuit::Spades,
            index: CardIndex::K,
        });

        hand.merge(&mut other);

        assert_eq!(other.count(), 0);
        assert_eq!(hand.count(), 3);
    }
}
