use std::sync::Mutex;

use rocket::{
    http::Status,
    serde::{json::Json, Serialize},
    Build, Rocket, State,
};

use crate::{
    blackjack::{
        blackjack_game::BlackjackGame, gamestate::GameState, play_moves::PlayMoves, player::Player,
    },
    cards::{card::Card, hand::Hand},
};

#[get("/state")]
fn get_state(state: &State<MyState>) -> Json<ExtendedGameState> {
    Json(gamestate_as_json(&state.game.lock().unwrap()))
}

#[post("/state", data = "<request>", format = "application/json")]
fn post_move(
    state: &State<MyState>,
    request: Json<PlayMoves>,
) -> Result<Json<ExtendedGameState>, Status> {
    let action = request.into_inner();

    if state.game.lock().unwrap().play(action) {
        Ok(Json(gamestate_as_json(&state.game.lock().unwrap())))
    } else {
        Err(Status::BadRequest)
    }
}

struct MyState {
    game: Mutex<BlackjackGame>,
}

#[derive(Serialize)]
#[serde(tag = "type", rename_all = "camelCase", rename_all_fields = "camelCase")]
enum ExtendedGameState {
    Over {
        dealer_hand: Hand,
        players: Vec<Player>,
    },
    Playing {
        dealer_upcard: Card,
        player_turn: usize,
        hand_turn: usize,
        players: Vec<Player>,
    },
    Starting,
}

fn gamestate_as_json(game: &BlackjackGame) -> ExtendedGameState {
    match game.get_state() {
        GameState::Starting => ExtendedGameState::Starting,
        GameState::Over => ExtendedGameState::Over {
            dealer_hand: game.get_dealer_hand().clone(),
            players: game.get_players().clone(),
        },
        GameState::PlayerTurn {
            player_index,
            hand_index,
        } => ExtendedGameState::Playing {
            dealer_upcard: *game.get_dealer_upcard().unwrap(),
            player_turn: *player_index,
            hand_turn: *hand_index,
            players: game.get_players().clone(),
        },
    }
}

pub fn build() -> Rocket<Build> {
    let state = MyState {
        game: Mutex::new(BlackjackGame::new()),
    };

    rocket::build()
        .manage(state)
        .mount("/api", routes![get_state, post_move])
}
